{-------------------------------------------------------------------------------------
-
- The XQuery Interpreter
- Programmer: Leonidas Fegaras
- Email: fegaras@cse.uta.edu
- Web: http://lambda.uta.edu/
- Creation: 03/22/08, last update: 10/20/08
- 
- Copyright (c) 2008 by Leonidas Fegaras, the University of Texas at Arlington. All rights reserved.
- This material is provided as is, with absolutely no warranty expressed or implied.
- Any use is at your own risk. Permission is hereby granted to use or copy this program
- for any purpose, provided the above notices are retained on all copies.
-
--------------------------------------------------------------------------------------}


{-# OPTIONS_GHC -fth -fglasgow-exts #-}


module Text.XML.HXQ.Interpreter where

import Control.Monad
import List(sortBy)
import XMLParse(parseDocument)
import System.Console.Readline
import Text.XML.HXQ.Parser
import Text.XML.HXQ.XTree
import Text.XML.HXQ.Optimizer
import Text.XML.HXQ.Functions
import Text.XML.HXQ.Compiler
import Text.XML.HXQ.OptionalDB


-- system functions (=, concat, etc)
systemFunctions :: [(String,Int,[XSeq]->XSeq)]
systemFunctions = $(iFunctions)


-- XPath step functions (child, descendant, etc)
pathFunctions :: [(String,Tag->XTree->XSeq)]
pathFunctions = $(pFunctions)


-- run-time bindings of FLOWR variables
type Environment = [(String,XSeq)]


-- a user-defined function is (fname,parameters,body)
type Functions = [(String,[String],Ast)]


undefv1 = error "Undefined XQuery context (.)"
undefv2 = error "Undefined position()"
undefv3 = error "Undefined last()"



-- Each XPath predicate must calculate position() and last() from its input XSeq
-- if last() is used, then the evaluation is blocking (need to store the whole input XSeq)
applyPredicates :: [Ast] -> XSeq -> Bool -> Environment -> Functions -> XSeq
applyPredicates [] xs _ _ _ = xs
applyPredicates ((Aint n):preds) xs _ env fncs   -- shortcut that improves laziness
    = applyPredicates preds [xs !! (n-1)] True env fncs
applyPredicates (pred:preds) xs True env fncs    -- top-k like
    | maxPosition pathPosition pred > 0
    = applyPredicates (pred:preds) (take (maxPosition pathPosition pred) xs) False env fncs
applyPredicates (pred:preds) xs _ env fncs
    | containsLast pred         -- blocking: use only when last() is used in the predicate
    = let last = length xs
      in applyPredicates preds
             (foldir (\x i r -> case eval pred x i last "" env fncs of
                                  [XInt k] -> if k == i then x:r else r               -- indexing
                                  b -> if conditionTest b then x:r else r) [] xs 1) True env fncs
applyPredicates (pred:preds) xs _ env fncs
    = applyPredicates preds
          (foldir (\x i r -> case eval pred x i undefv3 "" env fncs of
                               [XInt k] -> if k == i then x:r else r               -- indexing
                               b -> if conditionTest b then x:r else r) [] xs 1) True env fncs


-- The XQuery interpreter
-- context: context node (XPath .)
-- position: the element position in the parent sequence (XPath position())
-- last: the length of the parent sequence (XPath last())
-- effective_axis: the XPath axis in /axis::tag(exp)
--        eg, the effective axis of //(A | B) is descendant
-- env: contains FLOWR variable bindings
-- fncs: user-defined functions
eval :: Ast -> XTree -> Int -> Int -> String -> Environment -> Functions -> XSeq
eval e context position last effective_axis env fncs
  = case e of
      Avar "." -> [ context ]
      Avar v -> findV v env
      Aint n -> [ XInt n ]
      Afloat n -> [ XFloat n ]
      Astring s -> [ XText s ]
      Ast "context" [v,Astring dp,body]
          -> foldr (\x r -> (eval body x position last dp env fncs)++r)
                   [] (eval v context position last effective_axis env fncs)
      Ast "call" [Avar "position"] -> [XInt position]
      Ast "call" [Avar "last"] -> [XInt last]
      Ast "step" (Avar "child":tag:Avar ".":preds)
          | effective_axis /= ""
          -> eval (Ast "step" (Avar effective_axis:tag:Avar ".":preds)) context position last "" env fncs
      Ast "step" (Avar "descendant_any":Ast "tags" tags:e:preds)
          -> let ts = map (\(Avar tag) -> tag) tags
             in foldr (\x r -> (applyPredicates preds (descendant_any_with_tagged_children ts x) True env fncs)++r)
                      [] (eval e context position last effective_axis env fncs)
      Ast "step" (Avar step:Astring tag:e:preds)
          -> let step_fnc = findV step pathFunctions
             in foldr (\x r -> (applyPredicates preds (step_fnc tag x) True env fncs)++r)
                      [] (eval e context position last effective_axis env fncs)
      Ast "filter" (e:preds)
          -> applyPredicates preds (eval e context position last effective_axis env fncs) True env fncs
      Ast "predicate" [condition,body]
          -> if conditionTest (eval condition undefv1 undefv2 undefv3 "" env fncs)
             then eval body context position last effective_axis env fncs
             else []
      Ast "append" args
          -> appendText (map (\x -> eval x context position last effective_axis env fncs) args)
      Ast "call" ((Avar fname):args)
          -> case filter (\(n,_,_) -> n == fname || ("fn:"++n) == fname) systemFunctions of
               [(_,len,f)] -> if (length args) == len
                              then f (map (\x -> eval x context position last effective_axis env fncs) args)
                              else error ("Wrong number of arguments in system call: "++fname)
               _ -> case filter (\(n,_,_) -> n == fname) fncs of
                      (_,params,body):_ -> if (length params) == (length args)
                                           then eval body context undefv2 undefv3 ""
                                                    ((zipWith (\p a -> (p,eval a context position last effective_axis env fncs))
                                                              params args)++env) fncs
                                           else error ("Wrong number of arguments in function call: "++fname)
                      _ -> error ("Undefined function: "++fname)
      Ast "construction" [Astring tag,id,parent,Ast "attributes" [],body]
          -> let [XText vid] = eval id context position last effective_axis env fncs
                 [vparent] = eval parent context position last effective_axis env fncs
             in [ XElem tag [] (read vid) vparent (eval body context position last effective_axis env fncs) ]
      Ast "construction" [tag,id,parent,Ast "attributes" al,body]
             -> let alc = map (\(Ast "pair" [a,v])
                                     -> let ac = eval a context position last effective_axis env fncs
                                            vc = eval v context position last effective_axis env fncs
                                        in (qName ac,showXS vc)) al
                    ct = eval tag context position last effective_axis env fncs
                    bc = eval body context position last effective_axis env fncs
                    [XText vid] = eval id context position last effective_axis env fncs
                    [vparent] = eval parent context position last effective_axis env fncs
                in [ XElem (qName ct) alc (read vid) vparent bc ]
      Ast "let" [Avar var,source,body]
          -> eval body context position last effective_axis
                  ((var,eval source context position last effective_axis env fncs):env) fncs
      Ast "for" [Avar var,Avar "$",source,body]      -- a for-loop without an index
          -> foldr (\a r -> (eval body a undefv2 undefv3 "" ((var,[a]):env) fncs)++r)
                   [] (eval source context position last effective_axis env fncs)
      Ast "for" [Avar var,Avar ivar,source,body]     -- a for-loop with an index
          -> let p = maxPosition (Avar ivar) body
                 ns = if p > 0              -- there is a top-k like restriction
                      then Ast "step" [source,Ast "call" [Avar "<=",pathPosition,Aint p]]
                      else source 
             in foldir (\a i r -> (eval body a i undefv3 "" ((var,[a]):(ivar,[XInt i]):env) fncs)++r)
                       [] (eval ns context position last effective_axis env fncs) 1
      Ast "sortTuple" (exp:orderBys)             -- prepare each FLWOR tuple for sorting
          -> let ee = eval exp context position last effective_axis env fncs
             in [ XElem "" [] 0 parent_error
                     (foldl (\r a -> r++[XElem "" [] 0 parent_error (text (eval a context position last effective_axis env fncs))])
                            [XElem "" [] 0 parent_error ee] orderBys) ]
      Ast "sort" (exp:ordList)                   -- blocking
          -> let ce = map (\(XElem _ _ _ _ xs) -> map (\(XElem _ _ _ _ ys) -> ys) xs)
                          (eval exp context position last effective_axis env fncs)
                 ordering = foldr (\(Avar ord) r (x:xs) (y:ys)
                                       -> case compareXSeqs (ord == "ascending") x y of
                                            EQ -> r xs ys
                                            o -> o)
                                  (\xs ys -> EQ) ordList
             in concatMap head (sortBy (\(_:xs) (_:ys) -> ordering xs ys) ce)
      _ -> error ("Illegal XQuery: "++(show e))


type Statements = [(String,Statement)]


-- The monadic applyPredicates that propagates IO state
applyPredicatesM :: [Ast] -> XSeq -> Bool -> Environment -> Functions -> Statements -> IO XSeq
applyPredicatesM [] xs _ _ _ _ = return $! xs
applyPredicatesM ((Aint n):preds) xs _ env fncs stmts   -- shortcut that improves laziness
    = applyPredicatesM preds [xs !! (n-1)] True env fncs stmts
applyPredicatesM (pred:preds) xs True env fncs stmts    -- top-k like
    | maxPosition pathPosition pred > 0
    = applyPredicatesM (pred:preds) (take (maxPosition pathPosition pred) xs) False env fncs stmts
applyPredicatesM (pred:preds) xs _ env fncs stmts
    | containsLast pred         -- blocking: use only when last() is used in the predicate
    = do let last = length xs
         vs <- foldir (\x i r -> do vs <- evalM pred x i last "" env fncs stmts
                                    s <- r
                                    return $! (if case vs of
                                                    [XInt k] -> k == i               -- indexing
                                                    b -> conditionTest b
                                               then x:s else s))
                      (return []) xs 1
         applyPredicatesM preds vs True env fncs stmts
applyPredicatesM (pred:preds) xs _ env fncs stmts
    = do vs <- foldir (\x i r -> do vs <- evalM pred x i undefv3 "" env fncs stmts
                                    s <- r
                                    return $! (if case vs of
                                                    [XInt k] -> k == i               -- indexing
                                                    b -> conditionTest b
                                               then x:s else s))
                      (return []) xs 1
         applyPredicatesM preds vs True env fncs stmts


-- The monadic XQuery interpreter; it is like eval but has plumbing to propagate IO state
evalM :: Ast -> XTree -> Int -> Int -> String -> Environment -> Functions -> Statements -> IO XSeq
evalM e context position last effective_axis env fncs stmts
  = case e of
      Avar "." -> return $! [ context ]
      Avar v -> return $! (findV v env)
      Aint n -> return $! [ XInt n ]
      Afloat n -> return $! [ XFloat n ]
      Astring s -> return $! [ XText s ]
      -- for non-IO XQuery, use the regular eval
      Ast "nonIO" [u] -> return $! (eval u context position last effective_axis env fncs)
      Ast "context" [v,Astring dp,body]
          -> do vs <- evalM v context position last effective_axis env fncs stmts
                foldr (\x r -> (liftM2 (++)) (evalM body x position last dp env fncs stmts) r)
                      (return []) vs
      Ast "call" [Avar "position"] -> return $! [XInt position]
      Ast "call" [Avar "last"] -> return $! [XInt last]
      Ast "step" (Avar "child":tag:Avar ".":preds)
          | effective_axis /= ""
          -> evalM (Ast "step" (Avar effective_axis:tag:Avar ".":preds)) context position last "" env fncs stmts
      Ast "step" (Avar "descendant_any":Ast "tags" tags:e:preds)
          -> do vs <- evalM e context position last effective_axis env fncs stmts
                let ts = map (\(Avar tag) -> tag) tags
                foldr (\x r -> (liftM2 (++)) (applyPredicatesM preds (descendant_any_with_tagged_children ts x) True env fncs stmts) r)
                      (return []) vs
      Ast "step" (Avar step:Astring tag:e:preds)
          -> let step_fnc = findV step pathFunctions
             in do vs <- evalM e context position last effective_axis env fncs stmts
                   foldr (\x r -> (liftM2 (++)) (applyPredicatesM preds (step_fnc tag x) True env fncs stmts) r)
                         (return []) vs
      Ast "filter" (e:preds)
          -> do vs <- evalM e context position last effective_axis env fncs stmts
                applyPredicatesM preds vs True env fncs stmts
      Ast "predicate" [condition,body]
          -> do eb <- evalM condition undefv1 undefv2 undefv3 "" env fncs stmts
                if conditionTest eb
                   then evalM body context position last effective_axis env fncs stmts
                   else return []
      Ast "executeSQL" [Avar var,args]
          -> do as <- evalM args context position last effective_axis env fncs stmts
                executeSQL (findV var stmts) as
      Ast "call" [Avar nm,c,t,e]     -- this is the only lazy function
          | elem nm ["if","fn:if"]
          -> do ce <- evalM c context position last effective_axis env fncs stmts
                evalM (if conditionTest ce then t else e) context position last effective_axis env fncs stmts
      Ast "append" args
          -> (liftM appendText) (mapM (\x -> evalM x context position last effective_axis env fncs stmts) args)
      Ast "call" ((Avar fname):args)        -- Note: strict function application
          -> case filter (\(n,_,_) -> n == fname || ("fn:"++n) == fname) systemFunctions of
               [(_,len,f)] -> if (length args) == len
                              then (liftM f) (mapM (\x -> evalM x context position last effective_axis env fncs stmts) args)
                              else error ("Wrong number of arguments in system call: "++fname)
               _ -> case filter (\(n,_,_) -> n == fname) fncs of
                      (_,params,body):_ -> if (length params) == (length args)
                                           then do vs <- mapM (\a -> evalM a context position last effective_axis env fncs stmts) args
                                                   evalM body context undefv2 undefv3 ""
                                                             ((zipWith (\p a -> (p,a)) params vs)++env) fncs stmts
                                           else error ("Wrong number of arguments in function call: "++fname)
                      _ -> error ("Undefined function: "++fname)
      Ast "construction" [Astring tag,id,parent,Ast "attributes" [],body]
          -> do b <- evalM body context position last effective_axis env fncs stmts
                [XText vid] <- evalM id context position last effective_axis env fncs stmts
                [vparent] <- evalM parent context position last effective_axis env fncs stmts
                return $! [ XElem tag [] (read vid) vparent b ]
      Ast "construction" [tag,id,parent,Ast "attributes" al,body]
             -> do alc <- mapM (\(Ast "pair" [a,v])
                                     -> do ac <- evalM a context position last effective_axis env fncs stmts
                                           vc <- evalM v context position last effective_axis env fncs stmts
                                           return $! (qName ac,showXS vc)) al
                   ct <- evalM tag context position last effective_axis env fncs stmts
                   bc <- evalM body context position last effective_axis env fncs stmts
                   [XText vid] <- evalM id context position last effective_axis env fncs stmts
                   [vparent] <- evalM parent context position last effective_axis env fncs stmts
                   return $! [ XElem (qName ct) alc (read vid) vparent bc ]
      Ast "let" [Avar var,source,body]
          -> do s <- evalM source context position last effective_axis env fncs stmts
                evalM body context position last effective_axis ((var,s):env) fncs stmts
      Ast "for" [Avar var,Avar "$",source,body]      -- a for-loop without an index
          -> do vs <- evalM source context position last effective_axis env fncs stmts
                foldr (\a r -> (liftM2 (++)) (evalM body a undefv2 undefv3 "" ((var,[a]):env) fncs stmts) r)
                      (return []) vs
      Ast "for" [Avar var,Avar ivar,source,body]     -- a for-loop with an index
          -> do let p = maxPosition (Avar ivar) body
                    ns = if p > 0              -- there is a top-k like restriction
                            then Ast "step" [source,Ast "call" [Avar "<=",pathPosition,Aint p]]
                            else source 
                vs <- evalM ns context position last effective_axis env fncs stmts
                foldir (\a i r -> (liftM2 (++)) (evalM body a i undefv3 "" ((var,[a]):(ivar,[XInt i]):env) fncs stmts) r)
                       (return []) vs 1
      Ast "sortTuple" (exp:orderBys)             -- prepare each FLWOR tuple for sorting
          -> do vs <- evalM exp context position last effective_axis env fncs stmts
                os <- mapM (\a -> evalM a context position last effective_axis env fncs stmts) orderBys
                return $! [ XElem "" [] 0 parent_error (foldl (\r a -> r++[XElem "" [] 0 parent_error (text a)])
                                                              [XElem "" [] 0 parent_error vs] os) ]
      Ast "sort" (exp:ordList)                   -- blocking
          -> do vs <- evalM exp context position last effective_axis env fncs stmts
                let ce = map (\(XElem _ _ _ _ xs) -> map (\(XElem _ _ _ _ ys) -> ys) xs) vs
                    ordering = foldr (\(Avar ord) r (x:xs) (y:ys)
                                       -> case compareXSeqs (ord == "ascending") x y of
                                            EQ -> r xs ys
                                            o -> o)
                                  (\xs ys -> EQ) ordList
                return $! (concatMap head (sortBy (\(_:xs) (_:ys) -> ordering xs ys) ce))
      _ -> error ("Illegal XQuery: "++(show e))


-- evaluate from input continuously
evalInput :: (String -> Environment -> Functions -> IO(Environment,Functions)) -> Environment -> Functions -> IO ()
evalInput eval vs fs
    = do let oneline prompt = do line <- readline prompt
                                 case line of
                                   Nothing -> return "quit"
                                   Just t -> if t == ""
                                             then oneline prompt
                                             else return $! t
             readlines x = do line <- oneline ": "
                              if last line == '}'
                                 then return $! (x++" "++(init line))
                                 else if line == "quit"
                                      then return $! line
                                      else readlines (x++" "++line)
         line <- oneline "> "
         stmt <- if head line == '{'
                 then if last line == '}'
                      then return $! (init (tail line))
                      else readlines (tail line)
                 else return $! line
         if stmt == "quit"
            then putStrLn "Bye!"
            else do addHistory stmt
                    (nvs,nfs) <- eval (map (\c -> if c=='\"' then '\'' else c) stmt) vs fs
                    evalInput eval nvs nfs


evalQueryM :: [Ast] -> Environment -> Functions -> (String -> IO Statement) -> Bool -> IO (XSeq,Environment,Functions)
evalQueryM [] variables functions dbmapper verbose
    = return $! ([],variables,functions)
evalQueryM (query:xs) variables functions dbmapper verbose
    = case query of
        Ast "function" ((Avar f):b:args)
            -> evalQueryM xs variables ((f,map (\(Avar v) -> v) args,optimize b):functions) dbmapper verbose
        Ast "variable" [Avar v,u]
            -> do uv <- evalM (optimize u) undefv1 undefv2 undefv3 "" variables functions []
                  evalQueryM xs ((v,uv):variables) functions dbmapper verbose
        _ -> do let opt = optimize query
                    (ast,ns) = liftIOSources opt
                if verbose
                   then do putStrLn "Abstract Syntax Tree (AST):"
                           putStrLn (ppAst query)
                           putStrLn "Optimized AST:"
                           putStrLn (ppAst opt)
                           putStrLn "Result:"
                   else return ()
                env <- foldr (\(n,b,s) r -> case s of
                                              Avar m
                                                  -> do env <- r
                                                        return $! ((n,findV m env):env)
                                              Astring file
                                                  -> do doc <- readFile file
                                                        env <- r
                                                        return $! ((n,[materialize b (parseDocument doc)]):env)
                                              _ -> r)
                             (return []) ns
                stmts <- foldr (\(n,_,s) r -> case s of
                                                Ast "prepareSQL" [Astring sql]
                                                    -> do stmts <- r
                                                          t <- dbmapper sql
                                                          return $! ((n,t):stmts)
                                                _ -> r)
                               (return []) ns
                result <- evalM ast undefv1 undefv2 undefv3 "" (env++variables) functions stmts
                (rest,renv,rfuns) <- evalQueryM xs variables functions dbmapper verbose
                return $! (result++rest,renv,rfuns)


xqueryE :: String -> Environment -> Functions -> (String -> IO Statement) -> Bool -> IO (XSeq,Environment,Functions)
xqueryE query variables functions dbmapper verbose
    = evalQueryM (parse (scan query)) variables functions dbmapper verbose


-- | Evaluate the XQuery using the interpreter.
xquery :: String -> IO XSeq
xquery query = do (u,_,_) <- xqueryE query [] [] (\sql -> error "No database connectivity") False
                  return $! u


-- | Read an XQuery from a file and run it using the interpreter.
xfile :: String -> IO XSeq
xfile file = do query <- readFile file
                xquery query


-- | Evaluate the XQuery with database connectivity using the interpreter.
xqueryDB :: (IConnection conn) => String -> conn -> IO XSeq
xqueryDB query db = do (u,_,_) <- xqueryE query [] [] (prepareSQL db) False
                       return $! u


-- | Read an XQuery with database connectivity from a file and run it using the interpreter.
xfileDB :: (IConnection conn) => String -> conn -> IO XSeq
xfileDB file db = do query <- readFile file
                     xqueryDB query db
